@@ -1792,7 +1792,7 @@ class CustomTarget(Target):
         if 'build_always' in kwargs and 'build_always_stale' in kwargs:
             raise InvalidArguments('build_always and build_always_stale are mutually exclusive. Combine build_by_default and build_always_stale.')
         elif 'build_always' in kwargs:
-            mlog.warning('build_always is deprecated. Combine build_by_default and build_always_stale instead.')
+            mlog.deprecation('build_always is deprecated. Combine build_by_default and build_always_stale instead.')
             if 'build_by_default' not in kwargs:
                 self.build_by_default = kwargs['build_always']
             self.build_always_stale = kwargs['build_always']
