@@ -25,10 +25,13 @@ __all__ = [
     'ICC_STANDARD',
     'ICC_WIN',
 
+    'all_languages',
     'base_options',
-    'clike_langs',
+    'clib_langs',
+    'clink_langs',
     'c_suffixes',
     'cpp_suffixes',
+    'get_macos_dylib_install_name',
     'get_base_compile_args',
     'get_base_link_args',
     'is_assembly',
@@ -39,8 +42,10 @@ __all__ = [
     'is_source',
     'lang_suffixes',
     'sanitizer_compile_args',
-    'sort_clike',
+    'sort_clink',
 
+    'ArmCCompiler',
+    'ArmCPPCompiler',
     'CCompiler',
     'ClangCCompiler',
     'ClangCompiler',
@@ -97,10 +102,13 @@ from .compilers import (
     ICC_OSX,
     ICC_WIN,
     ICC_STANDARD,
+    all_languages,
     base_options,
-    clike_langs,
+    clib_langs,
+    clink_langs,
     c_suffixes,
     cpp_suffixes,
+    get_macos_dylib_install_name,
     get_base_compile_args,
     get_base_link_args,
     is_header,
@@ -111,15 +119,16 @@ from .compilers import (
     is_library,
     lang_suffixes,
     sanitizer_compile_args,
-    sort_clike,
+    sort_clink,
     ClangCompiler,
     CompilerArgs,
     GnuCompiler,
     IntelCompiler,
 )
 from .c import (
-    ArmCCompiler,
     CCompiler,
+    ArmCCompiler,
+    ArmclangCCompiler,
     ClangCCompiler,
     GnuCCompiler,
     ElbrusCCompiler,
@@ -127,8 +136,9 @@ from .c import (
     VisualStudioCCompiler,
 )
 from .cpp import (
-    ArmCPPCompiler,
     CPPCompiler,
+    ArmCPPCompiler,
+    ArmclangCPPCompiler,
     ClangCPPCompiler,
     GnuCPPCompiler,
     ElbrusCPPCompiler,
